"""
stignore-agent

A basic flask API that provides a set of endpoints to:
* Work with content types (folders underneath the base)
* Manipulate each content types .stignore file
"""
import shutil

from flask import Flask, current_app, request, jsonify, send_from_directory

from stignore_agent.helpers import load_stignore_file, stignore_actions, load_actions


app = Flask("stignore-agent")


@app.after_request
def after_request(response):
    """Apply extra CORS header"""
    headers = response.headers
    headers["Access-Control-Allow-Origin"] = "*"
    headers["Access-Control-Allow-Methods"] = "*"
    headers["Access-Control-Allow-Headers"] = "*"
    return response


@app.route("/")
def info_page():
    """Basic info page for users discovering this through their browser"""
    return jsonify(
        {
            "ok": True,
            "msg": "Please visit https://github.com/dalmura/stignore-agent for more information",
        }
    )


@app.route("/favicon.ico")
def favicon():
    """Static favicon covering the above info page"""
    return send_from_directory("static/img", "favicon.ico")


@app.route("/api/v1/discover")
def list_content_types():
    """Returns all configured content type folders"""
    folders = current_app.config["folders"]

    return jsonify(
        {"ok": True, "content_types": list({"name": name} for name in folders.keys())}
    )


@app.route("/api/v1/<content_type>/listing")
def content_type_listing(content_type: str):
    """
    Given a valid content type we return a listing of all folders underneath it
    Also respecting configured search depth
    """
    folders = current_app.config["folders"]

    content_folder = folders.get(content_type)

    if content_folder is None:
        return (
            jsonify({"ok": False, "msg": "Provided content_type is not monitored"}),
            400,
        )

    search_depth = content_folder.depth + 1

    if not content_folder.path.exists():
        return (
            jsonify({"ok": False, "msg": "Provided content_type does not exist"}),
            400,
        )

    folders = []

    search_glob = "/".join("*" * search_depth)

    for content in content_folder.path.glob(search_glob):
        if not content.is_dir():
            # We're only interested in folders
            continue

        if content.name.startswith(".st"):
            # We skip the syncthing specific folders
            continue

        folders.append(
            {
                "name": content.name,
                "size_megabytes": round(
                    sum(f.stat().st_size for f in content.glob("**/*") if f.is_file())
                    / 1024
                    / 1024,
                    2,
                ),
            }
        )

    return jsonify(
        {
            "ok": True,
            "folders": sorted(folders, key=lambda x: x["name"]),
        }
    )


@app.route("/api/v1/<content_type>/stignore")
def stignore_listing(content_type: str):
    """
    Given a valid content type we return the .stignore files contents
    """
    folders = current_app.config["folders"]

    content_folder = folders.get(content_type)

    if content_folder is None:
        return (
            jsonify({"ok": False, "msg": "Provided content_type is not monitored"}),
            400,
        )

    stignore = content_folder.path / ".stignore"

    if not stignore.exists():
        return (
            jsonify(
                {
                    "ok": False,
                    "msg": ".stignore doesn't exists for the provided content_type",
                }
            ),
            400,
        )

    return jsonify(
        {
            "ok": True,
            "entries": load_stignore_file(stignore),
        }
    )


@app.route("/api/v1/<content_type>/stignore", methods=["POST"])
def stignore_modification(content_type: str):
    """
    Apply a modification to the stignore file
    We do not perform any 'cleanup' associated with the modification
    Calling the flush endpoint will trigger that
    """
    folders = current_app.config["folders"]

    content_folder = folders.get(content_type)

    if content_folder is None:
        return (
            jsonify({"ok": False, "msg": "Provided content_type is not monitored"}),
            400,
        )

    stignore = content_folder.path / ".stignore"

    if not stignore.exists():
        return jsonify(
            {
                "ok": False,
                "msg": ".stignore doesn't exists for the provided content_type",
            }
        )

    entries = load_stignore_file(stignore)
    raw_entries = [e["raw"] for e in entries]

    # Insert the payload
    payload = request.get_json(force=True)

    if "actions" not in payload:
        return jsonify({"ok": False, "msg": "No provided actions"})

    actions = load_actions(payload["actions"])

    if not actions["ok"]:
        return jsonify(actions)

    for entry in actions["remove"]:
        if entry in raw_entries:
            raw_entries.remove(entry)

    for entry in actions["add"]:
        if entry not in raw_entries:
            raw_entries.append(entry)

    # Write out new stignore file
    raw_entries.sort()
    stignore.write_text("\n".join(raw_entries) + "\n")

    return jsonify({"ok": True, "msg": "Actions applied"})


@app.route("/api/v1/<content_type>/stignore/flush")
def stignore_flush_report(content_type: str):
    """
    Prepare a list of actions that would occur if a flush was to happen
    This is a fail safe for the user to verify what *would* happen
    """
    folders = current_app.config["folders"]

    content_folder = folders.get(content_type)

    if content_folder is None:
        return (
            jsonify({"ok": False, "msg": "Provided content_type is not monitored"}),
            400,
        )

    stignore = content_folder.path / ".stignore"

    if not stignore.exists():
        return (
            jsonify(
                {
                    "ok": False,
                    "msg": ".stignore doesn't exist for the provided content_type",
                }
            ),
            400,
        )

    entries = load_stignore_file(stignore)
    actions = stignore_actions(entries, content_folder.path)

    return jsonify(
        {
            "ok": True,
            "actions": actions,
        }
    )


@app.route("/api/v1/<content_type>/stignore/flush", methods=["POST"])
def stignore_flush_delete(content_type: str):
    # pylint: disable=too-many-branches
    """
    Flush the stignore file by performing all operations marked in it
    This is mainly used to clean up all folders we've marked to ignore
    """
    folders = current_app.config["folders"]

    content_folder = folders.get(content_type)

    if content_folder is None:
        return (
            jsonify({"ok": False, "msg": "Provided content_type is not monitored"}),
            400,
        )

    stignore = content_folder.path / ".stignore"

    if not stignore.exists():
        return (
            jsonify(
                {
                    "ok": False,
                    "msg": ".stignore doesn't exists for the provided content_type",
                }
            ),
            400,
        )

    payload = request.get_json(force=True)
    payload_actions = payload.get("actions")

    if not payload_actions:
        return (
            jsonify({"ok": False, "msg": "Missing 'actions' confirmation"}),
            400,
        )

    entries = load_stignore_file(stignore)
    actions = stignore_actions(entries, content_folder.path)

    if len(actions) != len(payload_actions):
        return (
            jsonify(
                {
                    "ok": False,
                    "msg": "Invalid actions payload validation (invalid length)",
                }
            ),
            400,
        )

    for i, (src, dst) in enumerate(zip(payload_actions, actions), start=1):
        if src.get("name") != dst.get("name"):
            valid = False
        elif src.get("path") != dst.get("path"):
            valid = False
        elif src.get("action") != dst.get("action"):
            valid = False
        elif src.get("size_megabytes") != dst.get("size_megabytes"):
            valid = False
        else:
            valid = True

        if not valid:
            return (
                jsonify(
                    {
                        "ok": False,
                        "msg": f"Invalid actions payload validation (item {i})",
                    }
                ),
                400,
            )

    for action in actions:
        if action["action"] != "delete":
            continue

        shutil.rmtree(action["path"])

    return jsonify(
        {
            "ok": True,
            "actions": actions,
        }
    )



"""
stignore-agent helpers

Various helper functions to remove complexity from app views
"""
from types import SimpleNamespace
from pathlib import Path


def parse_config(config):
    """
    Takes a basic config object and returns the transformed one the app requires
    """
    return {
        "base_folder": Path(config["base_folder"]),
        "folders": {
            folder["name"]: SimpleNamespace(
                path=Path(config["base_folder"]) / folder["name"],
                depth=folder.get("depth", 0),
            )
            for folder in config["folders"]
        },
    }


def load_stignore_file(filename, sort=True):
    """
    Loads a provided stignore filename
    Parses it into a list of entry objects and returns them
    """
    entries = []

    with open(filename, "rt", encoding="utf-8") as stignore_file:
        for line in stignore_file:
            if line.endswith("\n"):
                line = line[:-1]

            if line.startswith("!"):
                ignore_type = "keep"
                name = line[1:]
            else:
                ignore_type = "ignore"
                name = line

            # We want to drop the trailing slash
            # As this means 'the contents of the folder but not the folder itself
            # And we want the folder itself included in this decision
            if name.endswith("/"):
                name = name[:-1]

            entries.append(
                {
                    "raw": line,
                    "name": name,
                    "ignore_type": ignore_type,
                }
            )

    if sort:
        entries = sorted(entries, key=lambda x: x["raw"])

    return entries


def stignore_actions(entries, content_folder, include_size=True):
    """
    Takes a list of stignore entities
    Returns a list of actions to align the entities to what appears on disk
    """
    actions = []

    for entry in entries:
        if entry["ignore_type"] != "ignore":
            # We're only looking for entries that could result in cleaning up
            continue

        entry_path = content_folder / entry["name"]

        if not entry_path.exists():
            # Only looking for entry_path's that exist
            continue

        action = {
            "name": str(entry_path.name),
            "path": str(entry_path),
            "action": "delete",
        }

        if include_size:
            size_bytes = sum(
                f.stat().st_size for f in entry_path.glob("**/*") if f.is_file()
            )
            action["size_megabytes"] = size_bytes / 1024 / 1024

        actions.append(action)

    return actions


def load_actions(actions):
    """
    Parses a list of provided entity actions
    Returns a list of raw entity lines to add, and another to remove
    """
    parsed = {
        "ok": True,
        "add": [],
        "remove": [],
    }

    for action in actions:
        if action.get("action") not in ("add", "remove"):
            return {"ok": False, "msg": "Payload action is invalid"}

        if action.get("ignore_type") not in ("keep", "ignore"):
            return {"ok": False, "msg": "Payload ignore_type is invalid"}

        if action.get("ignore_type") == "keep":
            new_entry = f"!{action['name']}"
        else:
            new_entry = action["name"]

        if not new_entry.endswith("/"):
            new_entry = f"{new_entry}/"

        if action.get("action") == "add":
            parsed["add"].append(new_entry)
        elif action.get("action") == "remove":
            parsed["remove"].append(new_entry)
        else:
            return {"ok": False, "msg": "Payload action is invalid"}

    return parsed
